import typing
from ..environment import Environment
from ..keyword import Keyword
from ..error import OperatorNotSupportedError


class BinaryOperator:
    def __init__(self, op: Keyword, left: typing.Any, right: typing.Any):
        self.op = op
        self.left = left
        self.right = right

    def evaluate(self, env: Environment) -> typing.Any:
        l = env.evaluate(self.left)
        r = env.evaluate(self.right)
        if self.op == Keyword.ADD:
            return l + r
        if self.op == Keyword.SUBTRACT:
            return l - r
        if self.op == Keyword.MULTIPLY:
            return l * r
        if self.op == Keyword.DIVIDE:
            return l / r
        if self.op == Keyword.FLOOR_DIVIDE:
            return l // r
        if self.op == Keyword.MODULO:
            return l % r
        if self.op == Keyword.POWER:
            return l**r

        if self.op == Keyword.EQUAL:
            return l == r
        if self.op == Keyword.NOT_EQUAL:
            return l != r
        if self.op == Keyword.LESS_THAN:
            return l < r
        if self.op == Keyword.LESS_THAN_EQUAL:
            return l <= r
        if self.op == Keyword.GREATER_THAN:
            return l > r
        if self.op == Keyword.GREATER_THAN_EQUAL:
            return l >= r

        if self.op == Keyword.AND:
            return l and r
        if self.op == Keyword.OR:
            return l or r

        if self.op == Keyword.IN:
            return l in r
        raise OperatorNotSupportedError(self.op)


class UnaryOperator:
    def __init__(self, op: Keyword, operand: typing.Any) -> None:
        self.op = op
        self.operand = operand

    def evaluate(self, env: Environment) -> typing.Any:
        value = env.evaluate(self.operand)
        if self.op == Keyword.NEGATIVE:
            return -value
        if self.op == Keyword.NOT:
            return not value
        raise OperatorNotSupportedError(self.op)



import enum
import typing
from .address import Address
from .block_result import BlockResult
from .environment import Environment


class BlockKind(enum.Enum):
    IFS = 0
    IF_ELIF_ELSE = 1

    FOR = 2
    WHILE = 3

    CALL = 4

    CLASS = 5


class Block:
    def __init__(
        self,
        kind: BlockKind,
        addr: Address,
        enter: typing.Callable[[Environment], bool],
        exit: typing.Callable[[Environment], BlockResult],
    ):
        self.kind = kind
        self.addr = addr.clone()
        self.enter = enter
        self.exit = exit

    def will_enter(self, env: Environment):
        env.addr = self.addr.clone()
        if self.enter(env):
            env.addr.shift(1)
            env.blocks.append(self)

    def did_exit(self, env: Environment) -> BlockResult:
        env.blocks.pop()
        return self.exit(env)



import typing
from ..block import Block, BlockKind
from ..block_result import BlockResult
from .command import Command
from ..environment import Environment
from ..error import (
    InvalidBreakError,
    InvalidContinueError,
    ObjectNotIterableError,
)
from ..expression.assignable import Assignable


class For(Command):
    def __init__(
        self,
        vars: typing.Union[Assignable, tuple],
        iterable: typing.Any,
    ):
        self.vars = vars
        self.iterable = iterable

    def execute(self, env: Environment) -> None:
        try:
            value = env.evaluate(self.iterable)
            iterator = iter(value)
        except TypeError:
            raise ObjectNotIterableError(str(self.iterable))

        def enter(env: Environment) -> bool:
            try:
                value = next(iterator)
            except StopIteration:
                return False
            if isinstance(self.vars, tuple):
                for var, val in zip(self.vars, value):
                    var.assign(val, env)
                return True
            self.vars.assign(value, env)
            return True

        def exit(env: Environment) -> BlockResult:
            block.will_enter(env)
            return BlockResult.JUMP

        block = Block(BlockKind.FOR, env.addr, enter, exit)
        block.will_enter(env)


class While(Command):
    def __init__(self, condition: typing.Any):
        self.condition = condition

    def execute(self, env: Environment) -> None:
        def enter(env: Environment) -> bool:
            return env.evaluate(self.condition)

        def exit(env: Environment) -> BlockResult:
            block.will_enter(env)
            return BlockResult.JUMP

        block = Block(BlockKind.WHILE, env.addr, enter, exit)
        block.will_enter(env)


class Break(Command):
    def execute(self, env: Environment) -> None:
        while True:
            block = env.blocks.pop()
            if block.kind in (BlockKind.IFS, BlockKind.IF_ELIF_ELSE):
                env.addr.shift(-1)
                continue
            elif block.kind in (BlockKind.FOR, BlockKind.WHILE):
                env.addr.shift(-1)
                break
            else:
                raise InvalidBreakError()


class Continue(Command):
    def execute(self, env: Environment) -> None:
        while True:
            block = env.blocks.pop()
            if block.kind in (BlockKind.IFS, BlockKind.IF_ELIF_ELSE):
                env.addr.shift(-1)
                continue
            elif block.kind in (BlockKind.FOR, BlockKind.WHILE):
                block.will_enter(env)
                break
            else:
                raise InvalidContinueError()



import sys
import ast
import json
import traceback

VERSION = "0.0.3"

KEYWORD_COMMENT = "#"

# expressions
KEYWORD_ATTR = "attr"
KEYWORD_CALL = "call"
KEYWORD_COMMA = ","
KEYWORD_KWARG = "kwarg"
KEYWORD_SUBSCRIPT = "sub"
KEYWORD_TUPLE = "tuple"
KEYWORD_VAR = "var"

# commands
KEYWORD_ASSIGN = "="
KEYWORD_BREAK = "break"
KEYWORD_CLASS_DEF = "class"
KEYWORD_CONTINUE = "continue"
KEYWORD_ELIF = "elif"
KEYWORD_ELSE = "else"
KEYWORD_END = "end"
KEYWORD_EXPR_STMT = "expr"
KEYWORD_FOR = "for"
KEYWORD_FUNC_DEF = "def"
KEYWORD_IF = "if"
KEYWORD_IFS = "ifs"
KEYWORD_IMPORT = "import"
KEYWORD_PASS = "pass"
KEYWORD_RETURN = "return"
KEYWORD_WHILE = "while"


class Python2CalciumVisitor(ast.NodeVisitor):
    def __init__(self, indent_spaces="    "):
        super().__init__()
        self.lines = []
        self.indents = []
        self.keyword = KEYWORD_COMMENT
        self.count_of_nested_if = 0
        self.indent_spaces = indent_spaces
        self.indent_offset = len(indent_spaces)

    def get_indent(self, node):
        return node.col_offset // self.indent_offset + 1 + self.count_of_nested_if

    def output_command(self, indent, keyword, elements=[]):
        self.indents.append(indent)
        comment = []
        line = [indent, comment, keyword]
        line.extend(elements)
        self.lines.append(json.JSONEncoder(ensure_ascii=False).encode(line))

    def output_first_line(self):
        self.output_command(1, KEYWORD_COMMENT, [VERSION])

    def output_end_of_code(self):
        self.output_command(1, KEYWORD_END)

    def output_node(self, node, keyword, elements=[]):
        indent = self.get_indent(node)
        self.output_command(indent, keyword, elements)

    def get_call(self, node):
        return self.visit(node.func), self.get_arguments(node.args)

    def get_arguments(self, args):
        return [self.visit(arg) for arg in args]

    def output_elif_or_else(self, node, indent):
        if (
            hasattr(node.orelse[0], "test")
            and node.orelse[0].col_offset == node.col_offset + self.indent_offset
        ):
            # eg.
            # else:
            #     if condition:
            self.output_command(indent, KEYWORD_ELSE)
            for stmt in node.orelse:
                self.visit(stmt)
        elif hasattr(node.orelse[0], "test"):
            self.output_elif(node.orelse[0], indent)
        else:
            self.output_command(indent, KEYWORD_ELSE)
            for stmt in node.orelse:
                self.visit(stmt)

    def output_elif(self, node, indent):
        # Should not call output_node()
        self.output_command(indent, KEYWORD_ELIF, [self.visit(node.test)])
        for stmt in node.body:
            self.visit(stmt)
        if len(node.orelse) != 0:
            self.output_elif_or_else(node, indent)

    def get_bin_op(self, node, op, left, right):
        return [self.get_operator(op), self.visit(left), self.visit(right)]

    def get_operator(self, op):
        if isinstance(op, ast.Add):
            return "+"
        elif isinstance(op, ast.Sub):
            return "-"
        elif isinstance(op, ast.Mult):
            return "*"
        elif isinstance(op, ast.Pow):
            return "**"
        elif isinstance(op, ast.Div):
            return "/"
        elif isinstance(op, ast.FloorDiv):
            return "//"
        elif isinstance(op, ast.Mod):
            return "%"
        elif isinstance(op, ast.BitAnd):
            return "&"
        elif isinstance(op, ast.BitOr):
            return "|"
        elif isinstance(op, ast.BitXor):
            return "^"
        elif isinstance(op, ast.LShift):
            return "<<"
        elif isinstance(op, ast.RShift):
            return ">>"
        elif isinstance(op, ast.And):
            return "and"
        elif isinstance(op, ast.Or):
            return "or"
        elif isinstance(op, ast.Eq):
            return "=="
        elif isinstance(op, ast.NotEq):
            return "!="
        elif isinstance(op, ast.Lt):
            return "<"
        elif isinstance(op, ast.LtE):
            return "<="
        elif isinstance(op, ast.Gt):
            return ">"
        elif isinstance(op, ast.GtE):
            return ">="
        elif isinstance(op, ast.Is):
            return "is"
        elif isinstance(op, ast.IsNot):
            return "is not"
        elif isinstance(op, ast.In):
            return "in"
        elif isinstance(op, ast.NotIn):
            return "not in"

    # Visit
    def visit_Module(self, node):
        self.output_first_line()
        for stmt in node.body:
            self.visit(stmt)
        self.output_end_of_code()

    def visit_Import(self, node):
        self.output_node(node, KEYWORD_IMPORT, [node.names[0].name])

    def visit_FunctionDef(self, node):
        elems = [node.name, [arg.arg for arg in node.args.args]]
        self.output_node(node, KEYWORD_FUNC_DEF, elems)
        for stmt in node.body:
            self.visit(stmt)

    def visit_ClassDef(self, node):
        elems: list[str | None] = [node.name]
        if len(node.bases) > 0:
            elems.append(self.visit(node.bases[0]))
        else:
            elems.append(None)
        self.output_node(node, KEYWORD_CLASS_DEF, elems)
        for stmt in node.body:
            self.visit(stmt)

    def visit_Assign(self, node):
        elems = []
        if not isinstance(node.targets[0], ast.Tuple):
            elems.append(self.visit(node.targets[0]))
        else:
            lhs = [KEYWORD_COMMA]
            lhs.extend([self.visit(n) for n in node.targets[0].elts])
            elems.append(lhs)
        elems.append(self.visit(node.value))
        self.output_node(node, KEYWORD_ASSIGN, elems)

    def visit_AugAssign(self, node):
        if isinstance(node.op, ast.Add):
            keyword = "+="
        elif isinstance(node.op, ast.Sub):
            keyword = "-="
        elif isinstance(node.op, ast.Mult):
            keyword = "*="
        elems = [self.visit(node.target), self.visit(node.value)]
        self.output_node(node, keyword, elems)

    def visit_Tuple(self, node):
        elems = [KEYWORD_TUPLE]
        elems.extend([self.visit(e) for e in node.elts])
        return elems

    def visit_For(self, node):
        elems = []
        var_elem = self.visit(node.target)
        if isinstance(var_elem, list) and var_elem[0] == KEYWORD_TUPLE:
            var_elem[0] = KEYWORD_COMMA  # convert to reference in Calcium
        elems.append(var_elem)
        elems.append(self.visit(node.iter))
        self.output_node(node, KEYWORD_FOR, elems)
        for stmt in node.body:
            self.visit(stmt)

    def visit_While(self, node):
        elems = [self.visit(node.test)]
        self.output_node(node, KEYWORD_WHILE, elems)
        for stmt in node.body:
            self.visit(stmt)

    def visit_If(self, node):
        self.output_node(node, KEYWORD_IFS)
        self.count_of_nested_if += 1
        self.output_node(node, KEYWORD_IF, [self.visit(node.test)])
        for stmt in node.body:
            self.visit(stmt)
        if len(node.orelse) != 0:
            indent = self.get_indent(node)
            self.output_elif_or_else(node, indent)
        self.count_of_nested_if -= 1

    def visit_Pass(self, node):
        self.output_command(self.get_indent(node), KEYWORD_PASS)

    def visit_Return(self, node):
        elems = []
        if hasattr(node, "value"):
            if node.value is None:
                elems.append(None)
            else:
                elems.append(self.visit(node.value))
        else:
            elems.append(None)
        self.output_command(self.get_indent(node), KEYWORD_RETURN, elems)

    def visit_Break(self, node):
        self.output_command(self.get_indent(node), KEYWORD_BREAK)

    def visit_Continue(self, node):
        self.output_command(self.get_indent(node), KEYWORD_CONTINUE)

    def visit_Compare(self, node):
        return self.get_bin_op(node, node.ops[0], node.left, node.comparators[0])

    def visit_BinOp(self, node):
        return self.get_bin_op(node, node.op, node.left, node.right)

    def visit_BoolOp(self, node):
        if isinstance(node.op, ast.And):
            op = "and"
        else:
            op = "or"
        elems = [op, self.visit(node.values[0]), self.visit(node.values[1])]
        count = len(node.values)
        i = 2
        while i < count:
            elems = [op, elems, self.visit(node.values[i])]
            i += 1
        return elems

    def visit_UnaryOp(self, node):
        elems = []
        if isinstance(node.op, ast.Not):
            keyword = "not"
        elif isinstance(node.op, ast.USub):
            keyword = "-_"
            if isinstance(node.operand, ast.Num):
                # Return by a literal with - sign
                return -node.operand.n
        elif isinstance(node.op, ast.Invert):
            keyword = "~"
        # if keyword not exist, then error will be raised
        elems.append(keyword)
        elems.append(self.visit(node.operand))
        return elems

    def visit_List(self, node):
        # List must be nested
        return [[self.visit(e) for e in node.elts]]

    def visit_Dict(self, node):
        obj = {}
        for k, v in zip(node.keys, node.values):
            if k is not None:
                obj[self.visit(k)] = self.visit(v)
        return obj

    def visit_Num(self, node):
        return node.n

    def visit_Constant(self, node):
        return node.value

    def visit_Str(self, node):
        return node.s.replace("\n", "\\n")

    def visit_Name(self, node):
        return [KEYWORD_VAR, node.id]

    def visit_NameConstant(self, node):
        return node.value

    def visit_Attribute(self, node):
        attrs: list[str | list] = [node.attr]
        childnode = node.value
        while isinstance(childnode, ast.Attribute):
            attrs.insert(0, childnode.attr)
            childnode = childnode.value
        if isinstance(childnode, ast.Name):
            attrs.insert(0, [KEYWORD_VAR, childnode.id])
        else:
            attrs.insert(0, childnode.value)  # type: ignore
        attrs.insert(0, KEYWORD_ATTR)
        return attrs

    def visit_Subscript(self, node):
        value = self.visit(node.value)
        if isinstance(node.slice, ast.Slice):
            if node.slice.lower != None:
                lower = self.visit(node.slice.lower)
            else:
                lower = None
            if node.slice.upper != None:
                upper = self.visit(node.slice.upper)
            else:
                upper = None
            return [KEYWORD_SUBSCRIPT, value, lower, upper]
        if isinstance(node.slice, ast.Name):
            sub = self.visit(node.slice)
        elif isinstance(node.slice, ast.Constant):
            sub = node.slice.value
        return [KEYWORD_SUBSCRIPT, value, sub]

    def visit_Expr(self, node):
        value = self.visit(node.value)
        return self.output_node(node, KEYWORD_EXPR_STMT, [value])

    def visit_Call(self, node):
        elems: list[str | list] = [KEYWORD_CALL]
        func_ref, args = self.get_call(node)
        elems.append(func_ref)
        elems.append(args)
        if len(node.keywords) > 0:
            for kwd in node.keywords:
                kwarg = [KEYWORD_KWARG]
                kwarg.append(kwd.arg)  # type: ignore
                kwarg.append(self.visit(kwd.value))
                # append to the list already added to elems
                args.append(kwarg)

        return elems

    def generic_visit(self, node):
        super().generic_visit(node)


def convert(src):
    try:
        module_node = ast.parse(src)
        visitor = Python2CalciumVisitor()
        visitor.visit(module_node)
        lines = []
        for indent, line in zip(visitor.indents, visitor.lines):
            lines.append("{}{}".format("  " * indent, line))
        code = (",\n").join(lines)
        return "[\n{}\n]\n".format(code)
    except Exception:
        return traceback.format_exc()


if __name__ == "__main__":
    filename = sys.argv[1]
    with open(filename, encoding="utf-8") as fin:
        json_array = convert(fin.read())
        print(json_array)



import typing
from .expression.assignable import (
    Assignable,
    Variable,
    Attribute,
    Subscript,
)
from .expression.call import Call, KeywordArgument
from .expression.operator import UnaryOperator, BinaryOperator
from .command.assign import Assign
from .command.class_stmt import Class
from .command.command import Command
from .command.expr_stmt import ExprStmt
from .command.function import Def, Return
from .command.ifs import Ifs, If, Elif, Else
from .command.import_stmt import Import
from .command.loop import For, While, Break, Continue
from .command.pass_stmt import Comment, Pass, End
from .element import Element
from .index import Index
from .keyword import Keyword


class Parser:
    def __init__(self):
        pass

    def read(self, line: list[Element]) -> Command:
        kwd: Keyword = Keyword(line[Index.KEYWORD])
        parser_func = _table[kwd]
        cmd = parser_func(self, line)
        return cmd

    def read_args(self, args_list: list[Element]) -> list[typing.Any]:
        args = []
        for elem in args_list:
            arg = self.read_expr(elem)
            args.append(arg)
        return args

    def read_expr(self, obj: Element) -> typing.Any:
        if isinstance(obj, dict):
            parsed_dict = {}
            for key, value in obj.items():
                # key must be str type
                parsed_dict[key] = self.read_expr(value)
            return parsed_dict
        if not isinstance(obj, list):
            return obj
        # when obj is list, it should have at least one element
        if isinstance(obj[0], list):
            # a list literal
            parsed_list = []
            for elem in obj[0]:
                parsed_list.append(self.read_expr(elem))
            return parsed_list
        # calcium's expression with keyword in the first element of a list
        kwd = Keyword(obj[Index.EXPRESSION_KEYWORD])
        if kwd in (Keyword.VARIABLE, Keyword.ATTRIBUTE, Keyword.SUBSCRIPT):
            return self.read_assignable(obj)
        if kwd == Keyword.CALL:
            callee = self.read_assignable(obj[Index.CALL_CALLEE])
            args = self.read_args(obj[Index.CALL_ARGS])
            return Call(callee, args)  # type: ignore
        if kwd in (Keyword.TUPLE, Keyword.COMMA):
            return tuple(
                self.read_expr(elem) for elem in obj[Index.EXPRESSION_KEYWORD + 1 :]
            )
        if kwd == Keyword.KWARG:
            return KeywordArgument(
                obj[Index.KWARG_NAME], self.read_expr(obj[Index.KWARG_VALUE])
            )
        if kwd in (Keyword.NOT, Keyword.NEGATIVE):
            operand = self.read_expr(obj[Index.UNARY_OPERAND])
            return UnaryOperator(kwd, operand)
        # should be a binary operator
        left = self.read_expr(obj[Index.LEFT_OPERAND])
        right = self.read_expr(obj[Index.RIGHT_OPERAND])
        return BinaryOperator(kwd, left, right)

    def read_assignable(
        self, listobj: list[Element]
    ) -> typing.Union[Assignable, tuple]:
        kwd = Keyword(listobj[Index.EXPRESSION_KEYWORD])
        if kwd == Keyword.VARIABLE:
            name: str = listobj[Index.VAR_NAME]  # type: ignore
            return Variable(name)
        if kwd == Keyword.ATTRIBUTE:
            obj: typing.Union[Assignable, str] = self.read_expr(listobj[Index.ATTR_OBJECT])  # type: ignore
            properties = []
            # attributes can chain property names
            for i in range(Index.ATTR_NAME, len(listobj)):
                properties.append(listobj[i])
            return Attribute(obj, properties)
        if kwd == Keyword.SUBSCRIPT:
            obj: typing.Union[Assignable] = self.read_assignable(listobj[Index.SUBSCRIPT_OBJECT])  # type: ignore
            if len(listobj) == Index.SUBSCRIPT_INDEX + 1:
                index = self.read_expr(listobj[Index.SUBSCRIPT_INDEX])
                return Subscript(obj, index)
            if len(listobj) == Index.SUBSCRIPT_SLICE_STOP + 1:
                start = self.read_expr(listobj[Index.SUBSCRIPT_SLICE_START])
                stop = self.read_expr(listobj[Index.SUBSCRIPT_SLICE_STOP])
                return Subscript(obj, None, start, stop)
        if kwd == Keyword.COMMA:
            assignables = []
            for elem in listobj[Index.EXPRESSION_KEYWORD + 1 :]:
                assignables.append(self.read_assignable(elem))  # type: ignore
            return tuple(assignables)
        raise ValueError("Invalid keyword for expression")


_table: dict[Keyword, typing.Callable[[Parser, list[Element]], Command]] = {}


def _assign(parser: Parser, line: list[Element]) -> Assign:
    left = parser.read_assignable(line[Index.ASSIGN_LEFT])  # type: ignore
    right = parser.read_expr(line[Index.ASSIGN_RIGHT])
    return Assign(left, right)


def _make_compound_assign(
    parser: Parser, line: list[Element], keyword: Keyword
) -> Assign:
    lhs = parser.read_assignable(line[Index.ASSIGN_LEFT])  # type: ignore
    rhs = parser.read_expr(line[Index.ASSIGN_RIGHT])
    binop = BinaryOperator(keyword, lhs, rhs)
    return Assign(lhs, binop)


def _compound_add(parser: Parser, line: list[Element]) -> Assign:
    return _make_compound_assign(parser, line, Keyword.ADD)


def _compound_subtract(parser: Parser, line: list[Element]) -> Assign:
    return _make_compound_assign(parser, line, Keyword.SUBTRACT)


def _compound_multiply(parser: Parser, line: list[Element]) -> Assign:
    return _make_compound_assign(parser, line, Keyword.MULTIPLY)


def _ifs(parser: Parser, line: list[Element]) -> Command:
    return Ifs()


def _if(parser: Parser, line: list[Element]) -> Command:
    condition = parser.read_expr(line[Index.CONDITION])
    return If(condition)


def _elif(parser: Parser, line: list[Element]) -> Command:
    condition = parser.read_expr(line[Index.CONDITION])
    return Elif(condition)


def _else(parser: Parser, line: list[Element]) -> Command:
    return Else()


def _for(parser: Parser, line: list[Element]) -> Command:
    variables = parser.read_assignable(line[Index.FOR_VARIABLES])  # type: ignore
    iterable = parser.read_expr(line[Index.FOR_ITERABLE])
    return For(variables, iterable)


def _while(parser: Parser, line: list[Element]) -> Command:
    condition = parser.read_expr(line[Index.CONDITION])
    return While(condition)


def _break(parser: Parser, line: list[Element]) -> Command:
    return Break()


def _continue(parser: Parser, line: list[Element]) -> Command:
    return Continue()


def _def(parser: Parser, line: list[Element]) -> Command:
    name: str = line[Index.DEF_NAME]  # type: ignore
    args = parser.read_args(line[Index.DEF_PARAMETERS])  # type: ignore
    return Def(name, args)


def _return(parser: Parser, line: list[Element]) -> Command:
    if len(line) < Index.RETURN_VALUE + 1:
        # return without value
        return Return(None)
    expr = parser.read_expr(line[Index.RETURN_VALUE])
    return Return(expr)


def _class(parser: Parser, line: list[Element]) -> Command:
    name: str = line[Index.CLASS_NAME]  # type: ignore
    superclass = parser.read_expr(line[Index.CLASS_SUPERCLASS])
    return Class(name, superclass)


def _import(parser: Parser, line: list[Element]) -> Command:
    path: str = line[Index.IMPORT_PATH]  # type: ignore
    return Import(path)


def _expr_stmt(parser: Parser, line: list[Element]) -> Command:
    expr = parser.read_expr(line[Index.EXPR_STMT])
    return ExprStmt(expr)


def _comment(parser: Parser, line: list[Element]) -> Command:
    return Comment()


def _pass(parser: Parser, line: list[Element]) -> Command:
    return Pass()


def _end(parser: Parser, line: list[Element]) -> Command:
    return End()


_table[Keyword.ASSIGN] = _assign
_table[Keyword.COMPOUND_ADD] = _compound_add
_table[Keyword.COMPOUND_SUBTRACT] = _compound_subtract
_table[Keyword.COMPOUND_MULTIPLY] = _compound_multiply
_table[Keyword.IFS] = _ifs
_table[Keyword.IF] = _if
_table[Keyword.ELIF] = _elif
_table[Keyword.ELSE] = _else
_table[Keyword.FOR] = _for
_table[Keyword.WHILE] = _while
_table[Keyword.BREAK] = _break
_table[Keyword.CONTINUE] = _continue
_table[Keyword.DEF] = _def
_table[Keyword.RETURN] = _return
_table[Keyword.CLASS] = _class
_table[Keyword.IMPORT] = _import
_table[Keyword.EXPR_STMT] = _expr_stmt
_table[Keyword.COMMENT] = _comment
_table[Keyword.PASS] = _pass
_table[Keyword.END] = _end



import sys

sys.path.append("../src")

from calciumlang.runtime import Runtime, RuntimeResult

from calciumlang.tools.converter import convert

import json

s = """
t = input('test: ')
print(t)
print(len(input(len(input(input(t))))))
"""

if len(sys.argv) > 1:
    with open(sys.argv[1]) as fin:
        s = fin.read()

code = convert(s)
print(code)
r = Runtime(json.loads(code))
try:
    result = r.run()
except:
    print(r.env.code[r.env.addr.line])

while True:
    if result == RuntimeResult.PAUSED:
        result = r.resume("test:")
        continue
    elif result == RuntimeResult.EXECUTED:
        result = r.run()
    else:
        break



import typing
from ..expression.assignable import Assignable
from .command import Command
from ..environment import Environment


class Assign(Command):
    def __init__(self, lhs: typing.Union[Assignable, tuple], rhs: typing.Any):
        self.lhs = lhs
        self.rhs = rhs

    def execute(self, env: Environment):
        value = env.evaluate(self.rhs)
        if not isinstance(self.lhs, tuple):
            self.lhs.assign(value, env)
            return
        for lhs, val in zip(self.lhs, value):  # type: ignore
            lhs.assign(val, env)



import sys
import unittest

sys.path.append("..")
sys.path.append("../../src")

from calciumlang.error import (
    OutOfRangeError,
    NameNotFoundError,
    ObjectNotIterableError,
    ObjectNotCallableError,
)

from run_test import run_calcium


class TestErrors(unittest.TestCase):
    def test_out_of_range(self):
        with self.assertRaises(OutOfRangeError) as context:
            run_calcium("out_of_range.py")
        self.append_message(context)

    def test_name_not_found(self):
        with self.assertRaises(NameNotFoundError) as context:
            run_calcium("name_not_found.py")
        self.append_message(context)

    def test_object_not_iterable(self):
        with self.assertRaises(ObjectNotIterableError) as context:
            run_calcium("object_not_iterable.py")
        self.append_message(context)

    def test_object_not_callable(self):
        with self.assertRaises(ObjectNotCallableError) as context:
            run_calcium("object_not_callable.py")
        self.append_message(context)

    def append_message(self, context):
        TestErrors.message += str(context.exception) + "\n"

    @classmethod
    def tearDownClass(cls):
        print(cls.message)

    message = "\n"


unittest.main()



from .command import Command
from ..environment import Environment
from ..error import InvalidModuleNameError


class Import(Command):
    def __init__(self, path: str):
        self.path = path

    def execute(self, env: Environment) -> None:
        module_names = self.path.split(".")
        for name in module_names:
            if not name.isalnum():
                raise InvalidModuleNameError(name)
        exec(f"import {self.path}")
        module_name = module_names[0]
        env.context.define(module_name, eval(module_name))



class A:
    def __init__(self, n):
        self.n = n


class B(A):
    def __init__(self, n, m):
        s = super(B, self)
        s.__init__(n)
        self.m = m * n


class C:
    class D:
        x = 7


a = A(1)
b = B(3, 7)
print(isinstance(b, A))

a.n += 10
print(a.n)
print(b.m)

a.d = C.D
d = a.d()
d.m = C()
d.m.x = a.d.x
print(d.m.x)



from inspect import signature
import typing
from .assignable import Attribute, Variable
from ..command.class_stmt import Class
from ..command.function import Def
from ..environment import Environment
from ..error import ObjectNotCallableError
from ..label import InputCalled


class Call:
    def __init__(
        self,
        callee: typing.Union[Variable, Attribute],
        args: list[typing.Any],
    ) -> None:
        self.callee = callee
        self.args = args
        self.value: typing.Any = None
        self.is_called = False
        self.is_returned = False

    def evaluate(self, env: Environment) -> typing.Any:
        funcobj = self.callee.evaluate(env)
        kwargs = {
            arg.kwd: env.evaluate(arg.value)
            for arg in self.args
            if isinstance(arg, KeywordArgument)
        }
        evaluated_args = [
            env.evaluate(arg)
            for arg in self.args
            if not isinstance(arg, KeywordArgument)
        ]

        # built-in input() function requires the runtime to be paused
        if funcobj is input:
            if not self.is_called:
                self.is_called = True
                if len(evaluated_args) > 0:
                    env.prompt = evaluated_args[0]
                raise InputCalled()
            if not self.is_returned:
                self.is_returned = True
                self.value = env.returned_value
                env.prompt = ""
            return self.value
        if funcobj is super:
            if not self.is_called:
                this = env.context.lookup("self")
                klass = this.__class__
                self.is_called, self.is_returned = True, True
                self.value = super(klass, this)
            return self.value

        if not self.is_called:
            self.is_called = True
            if callable(funcobj) and funcobj.__module__ in (
                Class.__module__,
                Def.__module__,
            ):
                # user defined function
                sig = signature(funcobj)
                if "*args" in str(sig):
                    kwargs["is_called_by_library"] = False
            try:
                self.value = funcobj(*evaluated_args, **kwargs)
                self.is_returned = True  # built-ins also reach here
                return self.value
            except TypeError:
                raise ObjectNotCallableError(str(funcobj))
        if not self.is_returned:
            self.is_returned = True
            self.value = env.returned_value
            env.returned_value = None
        return self.value


class KeywordArgument:
    def __init__(self, kwd: str, value: typing.Any) -> None:
        self.kwd = kwd
        self.value = value



import typing
from ..block import Block, BlockKind
from ..block_result import BlockResult
from .command import Command
from ..environment import Environment


class Ifs(Command):
    def execute(self, env: Environment):
        def enter(env: Environment) -> bool:
            return True

        def exit(env: Environment) -> BlockResult:
            env.addr.shift(-1)
            return BlockResult.SHIFT

        block = Block(BlockKind.IFS, env.addr, enter, exit)
        block.will_enter(env)


def _execute_conditional_block(env: Environment) -> None:
    def enter(env: Environment) -> bool:
        return True

    def exit(env: Environment) -> BlockResult:
        env.addr.shift(-2)
        env.blocks.pop()
        return BlockResult.JUMP

    block = Block(BlockKind.IF_ELIF_ELSE, env.addr, enter, exit)
    block.will_enter(env)


class If(Command):
    def __init__(self, condition: typing.Any):
        self.condition = condition

    def execute(self, env: Environment):
        if env.evaluate(self.condition):
            _execute_conditional_block(env)


class Elif(If):
    pass


class Else(Command):
    def execute(self, env: Environment):
        _execute_conditional_block(env)



s = "test"
i = 0
n = len(s)
print(n)
while i < n:
    c = s[i]
    if c.startswith("e"):
        print("e")
    elif c.endswith("s"):
        print("s")
    elif "t" in c:
        if c.find("t") != -1:
            print(c)
    i += 1



import json
from inspect import signature


class Runtime:
    def __init__(self, code_list):
        if isinstance(code_list, str):
            code_list = json.loads(code_list)
        self.env = Environment(code_list)
        self.breakpoints = set()
        self.parser = Parser()
        self._calls = []
        self._inputcmd = None

    def resume(self, inputstr):
        self.env.returned_value = inputstr
        cmd = self._inputcmd
        self._inputcmd = None
        try:
            cmd.execute(self.env)
            self.env.skip_to_next_line()
        except FunctionCalled:
            caller_addr = self.env.address.clone()
            self._calls.append(CallingCommand(caller_addr, cmd))
        except InputCalled:
            self._inputcmd = cmd
            return RESULT_PAUSED
        except:
            raise
        return RESULT_EXECUTED

    def run(self):
        while True:
            result = self.step()
            if result == RESULT_EXECUTED:
                continue
            else:
                return result

    def step(self):
        last_index = len(self.env.code) - 1
        if self.env.address.indent == 0:
            end_of_code = self.parser.read(self.env.code[last_index])
            if not isinstance(end_of_code, End):
                raise InvalidEnd()
            return RESULT_TERMINATED
        else:
            if self.env.address.line >= last_index:
                return RESULT_TERMINATED
            line = self.env.code[self.env.address.line]
            cmd = self.parser.read(line)

            caller_addr = self.env.address.clone()
            if len(self._calls) > 0 and caller_addr.is_at(self._calls[-1].address):
                lastcmd = self._calls.pop()
                cmd = lastcmd.command

            try:
                cmd.execute(self.env)
            except FunctionCalled:
                self._calls.append(CallingCommand(caller_addr, cmd))
            except InputCalled:
                self._inputcmd = cmd
                return RESULT_PAUSED
            except:
                raise

            if isinstance(cmd, End):
                return RESULT_TERMINATED

            self.env.skip_to_next_line()
            next_line = self.env.code[self.env.address.line]
            kw = next_line[INDEX_KEYWORD]
            while kw == KEYWORD_COMMENT or kw == KEYWORD_IFS:
                cmd = self.parser.read(next_line)
                cmd.execute(self.env)
                self.env.skip_to_next_line()
                next_line = self.env.code[self.env.address.line]
                kw = next_line[INDEX_KEYWORD]

            if self.env.address.line in self.breakpoints:
                return RESULT_BREAKPOINT
            else:
                return RESULT_EXECUTED


# Keyword
KEYWORD_ASSIGN = "="
KEYWORD_ADD = "+"
KEYWORD_SUBTRACT = "-"
KEYWORD_MULTIPLICATE = "*"
KEYWORD_EXPONENTIATE = "**"
KEYWORD_DIVIDE = "/"
KEYWORD_FLOOR_DIVIDE = "//"
KEYWORD_REMAIN = "%"

KEYWORD_COMPOUND_ADD = "+="
KEYWORD_COMPOUND_SUBTRACT = "-="
KEYWORD_COMPOUND_MULTIPLICATE = "*="

KEYWORD_EQUAL = "=="
KEYWORD_NOT_EQUAL = "!="
KEYWORD_LESS_THAN = "<"
KEYWORD_LESS_THAN_OR_EQUAL = "<="
KEYWORD_GREATER_THAN = ">"
KEYWORD_GREATER_THAN_OR_EQUAL = ">="
KEYWORD_AND = "and"
KEYWORD_OR = "or"
KEYWORD_IS = "is"
KEYWORD_IS_NOT = "is not"
KEYWORD_IN = "in"
KEYWORD_NOT_IN = "not in"

KEYWORD_BIT_AND = "&"
KEYWORD_BIT_OR = "|"
KEYWORD_BIT_XOR = "^"
KEYWORD_LEFT_SHIFT = "<<"
KEYWORD_RIGHT_SHIFT = ">>"

KEYWORD_NOT = "not"
KEYWORD_NEGATIVE = "-_"
KEYWORD_BIT_NOT = "~"

KEYWORD_IFS = "ifs"
KEYWORD_IF = "if"
KEYWORD_ELIF = "elif"
KEYWORD_ELSE = "else"

KEYWORD_FOR = "for"
KEYWORD_WHILE = "while"
KEYWORD_BREAK = "break"
KEYWORD_CONTINUE = "continue"

KEYWORD_DEF = "def"
KEYWORD_RETURN = "return"

KEYWORD_CLASS = "class"

KEYWORD_IMPORT = "import"

KEYWORD_EXPR_STMT = "expr"

KEYWORD_TRY = "try"
KEYWORD_EXCEPT = "except"
KEYWORD_RAISE = "raise"

KEYWORD_VAR = "var"
KEYWORD_ATTR = "attr"
KEYWORD_SUBSCRIPT = "sub"
KEYWORD_CALL = "call"

KEYWORD_COMMA = ","
KEYWORD_KWARG = "kwarg"
KEYWORD_TUPLE = "tuple"

KEYWORD_COMMENT = "#"
KEYWORD_PASS = "pass"
KEYWORD_END = "end"

# Index
INDEX_INDENT = 0
INDEX_OPTIONS = 1
INDEX_KEYWORD = 2

INDEX_ASSIGN_LHS = 3  # Left Hand Side
INDEX_ASSIGN_RHS = 4  # Right Hand Side

INDEX_CONDITION = 3

INDEX_FOR_VARIABLES = 3
INDEX_FOR_ITERABLE = 4

INDEX_DEF_NAME = 3
INDEX_DEF_PARAMETERS = 4

INDEX_RETURN_VALUE = 3

INDEX_CLASS_NAME = 3
INDEX_CLASS_SUPERCLASS = 4

INDEX_IMPORT_NAME = 3

INDEX_EXPR_STMT = 3

INDEX_EXCEPT_TYPE = 3
INDEX_EXCEPT_NAME = 4

INDEX_RAISE_EXCEPTION = 3

INDEX_EXPRESSION_KEYWORD = 0

INDEX_VAR_NAME = 1

INDEX_ATTR_OBJECT = 1
INDEX_ATTR_PROPERTIES = 2

INDEX_SUBSCRIPT_REFERENCE = 1
INDEX_SUBSCRIPT_INDEX = 2
INDEX_SUBSCRIPT_SLICE_START = 2
INDEX_SUBSCRIPT_SLICE_END = 3

INDEX_CALL_REFERENCE = 1
INDEX_CALL_ARGS = 2

INDEX_KWARG_KEYWORD = 1
INDEX_KWARG_VALUE = 2

INDEX_LEFT_OPERAND = 1
INDEX_RIGHT_OPERAND = 2

INDEX_UNARY_OPERAND = 1

# results of the runtime
RESULT_TERMINATED = 0
RESULT_EXECUTED = 1
RESULT_BREAKPOINT = 2
RESULT_EXCEPTION = 3
RESULT_PAUSED = 4

# results of the block
BLOCK_RESULT_JUMPPED = 0
BLOCK_RESULT_SHIFTED = 1


class Address:
    def __init__(self, indent, line, calls=0):
        self.indent = indent
        self.line = line
        self.calls = calls

    def clone(self):
        return Address(self.indent, self.line, self.calls)

    def is_at(self, addr):
        return (
            self.indent == addr.indent
            and self.line == addr.line
            and self.calls == addr.calls
        )

    def jump(self, addr):
        self.indent = addr.indent
        self.line = addr.line
        # calls is not affected

    def shift(self, delta_x, delta_y=0):
        self.indent += delta_x
        self.line += delta_y


# BlockKind
BLOCK_KIND_IFS = 0
BLOCK_KIND_IF_ELIF_ELSE = 1
BLOCK_KIND_FOR = 2
BLOCK_KIND_WHILE = 3
BLOCK_KIND_CALL = 4
BLOCK_KIND_CLASS_DEF = 5
BLOCK_KIND_TRY = 6
BLOCK_KIND_EXCEPT = 7


class Block:
    def __init__(self, kind, address, enter, exit):
        self.kind = kind
        self.address = address.clone()
        self.enter = enter
        self.exit = exit

    def will_enter(self, env):
        env.address = self.address.clone()
        if self.enter(env):
            env.address.shift(1)
            env.blocks.append(self)

    def did_exit(self, env):
        env.blocks.pop()
        return self.exit(env)


class CallingCommand:
    def __init__(self, address, command):
        self.address = address.clone()
        self.command = command


class Namespace:
    def __init__(self, parent, dictobj):
        self.parent = parent  # None is allowed
        self.dictobj = dictobj

    def register(self, name, obj):
        self.dictobj[name] = obj

    def lookup(self, name):
        try:
            return self.dictobj[name]
        except KeyError:
            if self.parent is None:
                raise NameError(name)
            return self.parent.lookup(name)

    def find_nesting_scope(self):
        scope = self
        while isinstance(scope, ClassScope):
            scope = scope.parent
        return scope


class ClassScope(Namespace):
    def create_attributes(self):
        return self.dictobj


class FuncScope(Namespace):
    pass


class GlobalScope(Namespace):
    def lookup(self, name):
        try:
            return super().lookup(name)
        except:
            try:
                if isinstance(__builtins__, dict):
                    return __builtins__[name]
                else:
                    return getattr(__builtins__, name)
            except KeyError:
                raise NameError(name)
            except:
                raise


class Variable:
    def __init__(self, name):
        self.name = name

    def assign(self, obj, env):
        env.context.register(self.name, obj)

    def evaluate(self, env):
        value = env.context.lookup(self.name)
        return value


class Attribute:
    def __init__(self, obj, properties):
        self.obj = obj
        self.properties = properties

    def assign(self, value, env):
        target = self._lookup(env)
        length = len(self.properties) - 1
        for i in range(length):
            target = getattr(target, self.properties[i])
        setattr(target, self.properties[length], value)

    def evaluate(self, env):
        target = self._lookup(env)
        for prop in self.properties:
            target = getattr(target, prop)
        return target

    def _lookup(self, env):
        if isinstance(self.obj, Variable) or isinstance(self.obj, Attribute):
            return self.obj.evaluate(env)
        else:
            return self.obj


class Subscript:
    def __init__(self, ref, index, start=None, end=None):
        self.ref = ref
        self.index = index
        self.start = start
        self.end = end

    def assign(self, value, env):
        obj = env.evaluate(self.ref)
        index = env.evaluate(self.index)
        start = env.evaluate(self.start)
        end = env.evaluate(self.end)
        if index is not None:
            obj[index] = value
        else:
            if start is None:
                if end is None:
                    obj[:] = value
                else:
                    obj[:end] = value
            else:
                if end is None:
                    obj[start:] = value
                else:
                    obj[start:end] = value

    def evaluate(self, env):
        obj = env.evaluate(self.ref)
        index = env.evaluate(self.index)
        start = env.evaluate(self.start)
        end = env.evaluate(self.end)
        if index is not None:
            return obj[index]
        else:
            if start is None:
                if end is None:
                    return obj[:]
                else:
                    return obj[:end]
            else:
                if self.end is None:
                    return obj[start:]
                else:
                    return obj[start:end]


class Call:
    def __init__(self, ref, args):
        self.ref = ref
        self.args = args
        self.value = None
        self.called = False
        self.returned = False

    def evaluate(self, env):
        callee = env.evaluate(self.ref)
        kwargs = {
            arg.keyword: env.evaluate(arg.value)
            for arg in self.args
            if isinstance(arg, KeywordArg)
        }
        evaluated_args = [
            env.evaluate(arg) for arg in self.args if not isinstance(arg, KeywordArg)
        ]

        # built-in input function requires runtime to be paused
        if callee is input:
            if not self.called:
                self.called = True
                if len(evaluated_args) > 0:
                    env.prompt = evaluated_args[0]
                raise InputCalled()
            else:
                if not self.returned:
                    self.returned = True
                    self.value = env.returned_value
                    env.returned_value = None
                    env.prompt = ""
                return self.value
        elif callee is super:
            if self.called:
                pass
            else:
                this = env.context.lookup("self")
                klass = this.__class__
                self.called, self.returned = True, True
                self.value = super(klass, this)
            return self.value

        if not self.called:
            self.called = True
            if callable(callee) and callee.__module__ == __name__:
                # user defined function
                sig = signature(callee)
                if "*args" in str(sig):
                    kwargs["is_called_by_library"] = False
            self.value = callee(*evaluated_args, **kwargs)
            self.returned = True  # built-ins reach here
            return self.value
        else:
            if not self.returned:
                self.returned = True
                self.value = env.returned_value
                env.returned_value = None
            return self.value


class KeywordArg:
    def __init__(self, keyword, value):
        self.keyword = keyword
        self.value = value


class BinaryOperation:
    def __init__(self, operator, left, right):
        self.operator = operator
        self.left = left
        self.right = right

    def evaluate(self, env):
        l = env.evaluate(self.left)
        r = env.evaluate(self.right)
        op = self.operator
        if op == KEYWORD_ADD:
            return l + r
        elif op == KEYWORD_SUBTRACT:
            return l - r
        elif op == KEYWORD_MULTIPLICATE:
            return l * r
        elif op == KEYWORD_EXPONENTIATE:
            return l**r
        elif op == KEYWORD_DIVIDE:
            return l / r
        elif op == KEYWORD_FLOOR_DIVIDE:
            return l // r
        elif op == KEYWORD_REMAIN:
            return l % r
        elif op == KEYWORD_EQUAL:
            return l == r
        elif op == KEYWORD_NOT_EQUAL:
            return l != r
        elif op == KEYWORD_LESS_THAN:
            return l < r
        elif op == KEYWORD_LESS_THAN_OR_EQUAL:
            return l <= r
        elif op == KEYWORD_GREATER_THAN:
            return l > r
        elif op == KEYWORD_GREATER_THAN_OR_EQUAL:
            return l >= r
        elif op == KEYWORD_AND:
            return l and r
        elif op == KEYWORD_OR:
            return l or r
        elif op == KEYWORD_IS:
            return l is r
        elif op == KEYWORD_IS_NOT:
            return l is not r
        elif op == KEYWORD_IN:
            return l in r
        elif op == KEYWORD_NOT_IN:
            return l not in r
        elif op == KEYWORD_BIT_AND:
            return l & r
        elif op == KEYWORD_BIT_OR:
            return l | r
        elif op == KEYWORD_BIT_XOR:
            return l ^ r
        elif op == KEYWORD_LEFT_SHIFT:
            return l << r
        elif op == KEYWORD_RIGHT_SHIFT:
            return l >> r


class UnaryOperation:
    def __init__(self, operator, operand):
        self.operator = operator
        self.operand = operand

    def evaluate(self, env):
        v = env.evaluate(self.operand)
        op = self.operator
        if op == KEYWORD_NOT:
            return not v
        elif op == KEYWORD_NEGATIVE:
            return -v
        elif op == KEYWORD_BIT_NOT:
            return ~v


class Environment:
    def __init__(self, code_list):
        self.code = code_list
        self.address = Address(1, 0)  # (indent, line)
        self.blocks = []
        self.callstack = []
        self.exception = None

        self.global_context = GlobalScope(None, {})
        self.context = self.global_context

        self.prompt = ""
        self.returned_value = None

    def evaluate(self, obj):
        if (
            isinstance(obj, int)
            or isinstance(obj, float)
            or isinstance(obj, bool)
            or isinstance(obj, str)
            or callable(obj)  # functions or classes
            or obj is None
        ):
            return obj
        elif isinstance(obj, list):
            return [self.evaluate(e) for e in obj]
        elif isinstance(obj, dict):
            return {k: self.evaluate(obj[k]) for k in obj.keys()}
        elif isinstance(obj, tuple):
            return tuple(self.evaluate(e) for e in obj)
        elif isinstance(obj, set):
            return {self.evaluate(e) for e in obj}
        else:
            return obj.evaluate(self)

    def skip_to_next_line(self):
        try:
            while True:
                next_line_index = self.address.line + 1
                try:
                    while True:
                        next_line = self.code[next_line_index]
                        next_indent = next_line[INDEX_INDENT]
                        delta = self.address.indent - next_indent
                        if delta >= 0:
                            for _ in range(delta):
                                block = self.blocks[-1]
                                result = block.did_exit(self)
                                if result == BLOCK_RESULT_JUMPPED:
                                    raise BreakInnerLoop()
                            raise BreakOuterLoop()
                        else:
                            next_line_index += 1
                            continue
                except BreakInnerLoop:
                    pass
                except:
                    raise
        except BreakOuterLoop:
            pass
        except:
            raise
        self.address.line = next_line_index


# commands
class Assign:
    def __init__(self, lhs, rhs):
        self.lhs = lhs
        self.rhs = rhs

    def execute(self, env):
        value = env.evaluate(self.rhs)
        if isinstance(self.lhs, tuple):
            for lhs, val in zip(self.lhs, value):
                lhs.assign(val, env)
        else:
            self.lhs.assign(value, env)


class Ifs:
    def execute(self, env):
        def enter(env):
            return True

        def exit(env):
            env.address.shift(-1)
            return BLOCK_RESULT_SHIFTED

        block = Block(BLOCK_KIND_IFS, env.address, enter, exit)
        block.will_enter(env)


def _execute_conditional_block(env):
    def enter(env):
        return True

    def exit(env):
        env.address.shift(-2)
        env.blocks.pop()
        return BLOCK_RESULT_JUMPPED

    block = Block(BLOCK_KIND_IF_ELIF_ELSE, env.address, enter, exit)
    block.will_enter(env)


class If:
    def __init__(self, condition):
        self.condition = condition

    def execute(self, env):
        if env.evaluate(self.condition):
            _execute_conditional_block(env)


class Elif(If):
    pass


class Else:
    def execute(self, env):
        _execute_conditional_block(env)


class For:
    def __init__(self, vars, iterable):
        self.vars = vars
        self.iterable = iterable

    def execute(self, env):
        iterator = iter(env.evaluate(self.iterable))

        def enter(env):
            try:
                next_value = next(iterator)
                if isinstance(self.vars, tuple):
                    for var, value in zip(self.vars, next_value):
                        var.assign(value, env)
                else:
                    self.vars.assign(next_value, env)
                return True
            except StopIteration:
                return False
            except:
                raise

        def exit(env):
            block.will_enter(env)
            return BLOCK_RESULT_JUMPPED

        block = Block(BLOCK_KIND_FOR, env.address, enter, exit)
        block.will_enter(env)


class While:
    def __init__(self, condition):
        self.condition = condition

    def execute(self, env):
        def enter(env):
            return env.evaluate(self.condition)

        def exit(env):
            block.will_enter(env)
            return BLOCK_RESULT_JUMPPED

        block = Block(BLOCK_KIND_WHILE, env.address, enter, exit)
        block.will_enter(env)


class Break:
    def execute(self, env):
        while True:
            block = env.blocks.pop()
            if (
                block.kind == BLOCK_KIND_IFS
                or block.kind == BLOCK_KIND_IF_ELIF_ELSE
                or block.kind == BLOCK_KIND_TRY
                or block.kind == BLOCK_KIND_EXCEPT
            ):
                env.address.shift(-1)
                continue
            elif block.kind == BLOCK_KIND_WHILE or block.kind == BLOCK_KIND_FOR:
                env.address.shift(-1)
                break
            else:
                raise InvalidBreak()


class Continue:
    def execute(self, env):
        while True:
            block = env.blocks.pop()
            if (
                block.kind == BLOCK_KIND_IFS
                or block.kind == BLOCK_KIND_IF_ELIF_ELSE
                or block.kind == BLOCK_KIND_TRY
                or block.kind == BLOCK_KIND_EXCEPT
            ):
                env.address.shift(-1)
                continue
            elif block.kind == BLOCK_KIND_WHILE or block.kind == BLOCK_KIND_FOR:
                block.will_enter(env)
                break
            else:
                raise InvalidContinue()


class Def:
    def __init__(self, name, params):
        self.name = name
        self.params = params

    def execute(self, env):
        defined_address = env.address.clone()
        nesting_scope = env.context.find_nesting_scope()
        is_classscope = isinstance(env.context, ClassScope)
        is_init = self.name == "__init__" and is_classscope

        def _func(*args, is_called_by_library=True):
            # could be called by standard libraries
            caller_addr = env.address.clone()
            local = FuncScope(
                nesting_scope,
                {param_name: arg for param_name, arg in zip(self.params, args)},
            )
            callee_addr = defined_address.clone()
            callee_addr.calls = caller_addr.calls + 1

            def enter(env):
                env.callstack.append(env.context)
                env.context = local
                return True

            has_exited = False

            def exit(env):
                env.address.jump(caller_addr)
                env.address.calls -= 1
                env.address.shift(0, -1)
                if is_init:
                    env.returned_value = env.context.lookup("self")
                env.context = env.callstack.pop()
                nonlocal has_exited
                has_exited = True
                return BLOCK_RESULT_JUMPPED

            block = Block(BLOCK_KIND_CALL, callee_addr, enter, exit)
            block.will_enter(env)
            if not is_called_by_library:
                raise FunctionCalled()
            else:
                parser = Parser()
                while not has_exited:
                    env.skip_to_next_line()
                    last_index = len(env.code) - 1
                    if env.address.indent == 0:
                        end_of_code = parser.read(env.code[last_index])
                        if not isinstance(end_of_code, End):
                            raise InvalidEnd()
                        break
                    else:
                        if env.address.line >= last_index:
                            break
                    line = env.code[env.address.line]
                    cmd = parser.read(line)
                    cmd.execute(env)

                env.skip_to_next_line()
                value = env.returned_value
                env.returned_value = None
                return value

        env.context.register(self.name, _func)


class Return:
    def __init__(self, expr):
        self.expr = expr

    def execute(self, env):
        env.returned_value = env.evaluate(self.expr)
        while True:
            block = env.blocks[-1]
            if block.kind == BLOCK_KIND_CALL:
                block.did_exit(env)
                return
            elif (
                block.kind == BLOCK_KIND_IFS
                or block.kind == BLOCK_KIND_IF_ELIF_ELSE
                or block.kind == BLOCK_KIND_TRY
                or block.kind == BLOCK_KIND_EXCEPT
                or block.kind == BLOCK_KIND_WHILE
                or block.kind == BLOCK_KIND_FOR
            ):
                env.blocks.pop()
                continue
            else:
                raise InvalidReturn()


class Class:
    def __init__(self, name, superclass):
        self.name = name
        self.superclass = superclass

    def execute(self, env):
        if self.superclass is None:
            superclass = object
        else:
            superclass = env.evaluate(self.superclass)

        def enter(env):
            parent = env.context.find_nesting_scope()
            env.context = ClassScope(parent, {})
            return True

        previous_context = env.context

        def exit(env):
            attributes = env.context.create_attributes()
            env.context = previous_context
            classtype = type(self.name, (superclass,), attributes)
            env.context.register(self.name, classtype)
            env.address.shift(-1)
            return BLOCK_RESULT_SHIFTED

        block = Block(BLOCK_KIND_CLASS_DEF, env.address, enter, exit)
        block.will_enter(env)


class Import:
    def __init__(self, name):
        self.name = name

    def execute(self, env):
        module_names = self.name.split(".")
        for name in module_names:
            if not name.isalnum():
                raise InvalidModuleName()
        exec("import {}".format(self.name))
        env.context.register(module_names[0], eval("{}".format(module_names[0])))


class ExprStmt:
    def __init__(self, value):
        self.value = value

    def execute(self, env):
        env.evaluate(self.value)


class Comment:
    def execute(self, env):
        pass  # do nothing


class Pass:
    def execute(self, env):
        pass  # do nothing


class End:
    def execute(self, env):
        pass  # do nothing


class Parser:
    def __init__(self):
        table = {}

        def _assign(line):
            lhs = self.read_ref(line[INDEX_ASSIGN_LHS])
            rhs = self.read_expr(line[INDEX_ASSIGN_RHS])
            return Assign(lhs, rhs)

        table[KEYWORD_ASSIGN] = _assign

        def _make_compound_assign(keyword, line):
            lhs = self.read_ref(line[INDEX_ASSIGN_LHS])
            rhs = self.read_expr(line[INDEX_ASSIGN_RHS])
            binop = BinaryOperation(keyword, lhs, rhs)
            return Assign(lhs, binop)

        def _compound_add(line):
            return _make_compound_assign(KEYWORD_ADD, line)

        table[KEYWORD_COMPOUND_ADD] = _compound_add

        def _compound_subtract(line):
            return _make_compound_assign(KEYWORD_SUBTRACT, line)

        table[KEYWORD_COMPOUND_SUBTRACT] = _compound_subtract

        def _compound_multiplicate(line):
            return _make_compound_assign(KEYWORD_MULTIPLICATE, line)

        table[KEYWORD_COMPOUND_MULTIPLICATE] = _compound_multiplicate

        def _ifs(line):
            return Ifs()

        table[KEYWORD_IFS] = _ifs

        def _if(line):
            condition = self.read_expr(line[INDEX_CONDITION])
            return If(condition)

        table[KEYWORD_IF] = _if

        def _elif(line):
            condition = self.read_expr(line[INDEX_CONDITION])
            return Elif(condition)

        table[KEYWORD_ELIF] = _elif

        def _else(line):
            return Else()

        table[KEYWORD_ELSE] = _else

        def _while(line):
            condition = self.read_expr(line[INDEX_CONDITION])
            return While(condition)

        table[KEYWORD_WHILE] = _while

        def _for(line):
            vars = self.read_ref(line[INDEX_FOR_VARIABLES])
            iterable = self.read_expr(line[INDEX_FOR_ITERABLE])
            return For(vars, iterable)

        table[KEYWORD_FOR] = _for

        def _break(line):
            return Break()

        table[KEYWORD_BREAK] = _break

        def _continue(line):
            return Continue()

        table[KEYWORD_CONTINUE] = _continue

        def _func_def(line):
            name = line[INDEX_DEF_NAME]
            params = line[INDEX_DEF_PARAMETERS]
            return Def(name, params)

        table[KEYWORD_DEF] = _func_def

        def _return(line):
            if len(line) - 1 < INDEX_RETURN_VALUE:
                return Return(None)
            else:
                expr = self.read_expr(line[INDEX_RETURN_VALUE])
                return Return(expr)

        table[KEYWORD_RETURN] = _return

        def _class_def(line):
            name = line[INDEX_CLASS_NAME]
            superclass = self.read_expr(line[INDEX_CLASS_SUPERCLASS])
            return Class(name, superclass)

        table[KEYWORD_CLASS] = _class_def

        def _import(line):
            name = line[INDEX_IMPORT_NAME]
            return Import(name)

        table[KEYWORD_IMPORT] = _import

        def _expr_stmt(line):
            value = self.read_expr(line[INDEX_EXPR_STMT])
            return ExprStmt(value)

        table[KEYWORD_EXPR_STMT] = _expr_stmt

        def _comment(line):
            return Comment()

        table[KEYWORD_COMMENT] = _comment

        def _pass(line):
            return Pass()

        table[KEYWORD_PASS] = _pass

        def _end(line):
            return End()

        table[KEYWORD_END] = _end

        self.table = table

    def read(self, line):
        kw = line[INDEX_KEYWORD]
        parser = self.table[kw]
        cmd = parser(line)
        return cmd

    def read_args(self, args_list):
        parsed_args = []
        for elem in args_list:
            arg = self.read_expr(elem)
            parsed_args.append(arg)
        return parsed_args

    def read_expr(self, obj):
        if isinstance(obj, list):
            if isinstance(obj[0], list):
                parsed_list = []
                for elem in obj[0]:
                    value = self.read_expr(elem)
                    parsed_list.append(value)
                return parsed_list
            kw = obj[INDEX_EXPRESSION_KEYWORD]
            if (
                kw == KEYWORD_VAR
                or kw == KEYWORD_ATTR
                or kw == KEYWORD_SUBSCRIPT
                or kw == KEYWORD_CALL
            ):
                ref = self.read_ref(obj)
                return ref
            elif kw == KEYWORD_TUPLE:
                return tuple(
                    self.read_expr(elem) for elem in obj[INDEX_EXPRESSION_KEYWORD + 1 :]
                )
            elif kw == KEYWORD_KWARG:
                return KeywordArg(
                    obj[INDEX_KWARG_KEYWORD], self.read_expr(obj[INDEX_KWARG_VALUE])
                )
            elif kw == KEYWORD_NOT or kw == KEYWORD_NEGATIVE or kw == KEYWORD_BIT_NOT:
                operand = self.read_expr(obj[INDEX_UNARY_OPERAND])
                unary_op = UnaryOperation(kw, operand)
                return unary_op
            else:
                left = self.read_expr(obj[INDEX_LEFT_OPERAND])
                right = self.read_expr(obj[INDEX_RIGHT_OPERAND])
                bin_op = BinaryOperation(kw, left, right)
                return bin_op

        if isinstance(obj, dict):
            parsed_dict = {}
            for key in obj:
                parsed_dict[key] = self.read_expr(obj[key])
            return parsed_dict
        else:
            return obj

    def read_ref(self, listobj):
        kw = listobj[INDEX_EXPRESSION_KEYWORD]
        if kw == KEYWORD_VAR:
            name = listobj[INDEX_VAR_NAME]
            return Variable(name)
        elif kw == KEYWORD_ATTR:
            obj = self.read_expr(listobj[INDEX_ATTR_OBJECT])
            properties = []
            for i in range(INDEX_ATTR_PROPERTIES, len(listobj)):
                properties.append(listobj[i])
            return Attribute(obj, properties)
        elif kw == KEYWORD_SUBSCRIPT:
            ref = self.read_ref(listobj[INDEX_SUBSCRIPT_REFERENCE])
            if len(listobj) == 3:
                index = self.read_expr(listobj[INDEX_SUBSCRIPT_INDEX])
                return Subscript(ref, index)
            elif len(listobj) == 4:
                start = self.read_expr(listobj[INDEX_SUBSCRIPT_SLICE_START])
                end = self.read_expr(listobj[INDEX_SUBSCRIPT_SLICE_END])
                return Subscript(ref, None, start, end)
        elif kw == KEYWORD_CALL:
            ref = self.read_ref(listobj[INDEX_CALL_REFERENCE])
            args = self.read_args(listobj[INDEX_CALL_ARGS])
            return Call(ref, args)
        elif kw == KEYWORD_COMMA:
            return tuple(
                self.read_expr(e) for e in listobj[INDEX_EXPRESSION_KEYWORD + 1 :]
            )


class BreakInnerLoop(Exception):
    pass


class BreakOuterLoop(Exception):
    pass


class FunctionCalled(Exception):
    pass


class InputCalled(Exception):
    pass


class InvalidBreak(Exception):
    pass


class InvalidContinue(Exception):
    pass


class InvalidEnd(Exception):
    pass


class InvalidModuleName(Exception):
    pass


class InvalidReturn(Exception):
    pass



class BaseCalciumError(Exception):
    def __init__(self, message: str):
        super().__init__(message)


class AssignmentNotSupportedError(BaseCalciumError):
    def __init__(self, obj: str):
        super().__init__(f"cannot assign to {obj}")
        self.obj = obj


class InvalidBreakError(BaseCalciumError):
    def __init__(self):
        super().__init__("break statement not within loop")


class InvalidContinueError(BaseCalciumError):
    def __init__(self):
        super().__init__("continue statement not within loop")


class InvalidModuleNameError(BaseCalciumError):
    def __init__(self, name: str):
        super().__init__(f"module name {name} is invalid")
        self.name = name


class InvalidReturnError(BaseCalciumError):
    def __init__(self):
        super().__init__("return statement not within function")


class KeyNotContainedError(BaseCalciumError):
    def __init__(self, obj: str, key: str):
        super().__init__(f"key {key.replace('\n', '\\n')} not contained in {obj}")
        self.obj = obj
        self.key = key


class NameNotFoundError(BaseCalciumError):
    def __init__(self, name: str):
        super().__init__(f"Name {name} is not defined")
        self.name = name


class ObjectNotCallableError(BaseCalciumError):
    def __init__(self, obj: str):
        super().__init__(f"object {obj} is not callable")
        self.obj = obj


class ObjectNotIterableError(BaseCalciumError):
    def __init__(self, obj: str):
        super().__init__(f"object {obj} is not iterable")
        self.obj = obj


class OperatorNotSupportedError(BaseCalciumError):
    def __init__(self, op: str):
        super().__init__(f"operator {op} not supported")
        self.op = op


class OutOfRangeError(BaseCalciumError):
    def __init__(self, obj: str, index: str):
        super().__init__(f"index out of range in {obj} at {index}")
        self.obj = obj
        self.index = index



from .command import Command
from ..environment import Environment


class Comment(Command):
    def execute(self, env: Environment) -> None:
        pass  # do nothing


class Pass(Command):
    def execute(self, env: Environment) -> None:
        pass  # do nothing


class End(Command):
    def execute(self, env: Environment) -> None:
        pass  # do nothing



import typing
from .error import NameNotFoundError


class Namespace:
    def __init__(
        self,
        parent: typing.Union["Namespace", None],
        dictobj: dict[str, typing.Any],
    ) -> None:
        self.parent = parent  # global scope has no parent
        self.dictobj = dictobj

    def define(self, name: str, value: typing.Any) -> None:
        self.dictobj[name] = value

    def lookup(self, name: str) -> typing.Any:
        try:
            return self.dictobj[name]
        except KeyError:
            if self.parent is None:
                raise NameNotFoundError(name)
            else:
                return self.parent.lookup(name)

    def find_nesting_scope(self) -> "Namespace":
        scope: Namespace = self
        while isinstance(scope, ClassScope):
            scope = scope.parent  # type: ignore
        return scope


class ClassScope(Namespace):
    def create_attributes(self) -> dict[str, typing.Any]:
        return self.dictobj


class FuncScope(Namespace):
    pass


class GlobalScope(Namespace):
    def lookup(self, name: str) -> typing.Any:
        try:
            return super().lookup(name)
        except:
            try:
                if isinstance(__builtins__, dict):
                    return __builtins__[name]
                else:
                    return getattr(__builtins__, name)
            except KeyError:
                raise NameNotFoundError(name)
            except:
                raise



import typing
from ..block import Block, BlockKind
from ..block_result import BlockResult
from .command import Command
from ..environment import Environment
from ..error import InvalidReturnError
from ..label import FunctionCalled
from ..namespace import ClassScope, FuncScope


class Def(Command):
    def __init__(self, name: str, params: list[str]):
        self.name = name
        self.params = params

    def execute(self, env: Environment) -> None:
        defined_addr = env.addr.clone()
        nesting_scope = env.context.find_nesting_scope()
        is_classscope = isinstance(env.context, ClassScope)
        is_init = self.name == "__init__" and is_classscope

        def _func(*args, is_called_by_library=True):
            # could be called by standard library
            caller_addr = env.addr.clone()
            local = FuncScope(
                nesting_scope,
                {param_name: arg for param_name, arg in zip(self.params, args)},
            )
            callee_addr = defined_addr.clone()
            callee_addr.calls = caller_addr.calls + 1

            def enter(env: Environment) -> bool:
                env.callstack.append(env.context)
                env.context = local
                return True

            had_exited = False

            def exit(env: Environment) -> BlockResult:
                env.addr.jump(caller_addr)
                env.addr.calls -= 1
                env.addr.shift(0, -1)
                if is_init:
                    env.returned_value = env.context.lookup("self")
                env.context = env.callstack.pop()
                nonlocal had_exited
                had_exited = True
                return BlockResult.JUMP

            block = Block(BlockKind.CALL, callee_addr, enter, exit)
            block.will_enter(env)

            if not is_called_by_library:
                # controls the flow of the program
                raise FunctionCalled()

            from ..parser import Parser

            parser = Parser()
            while not had_exited:
                env.update_addr_to_next_command()
                last_index = len(env.code) - 1
                if env.addr.indent == 0:
                    break
                if env.addr.line >= last_index:
                    break
                line = env.code[env.addr.line]
                cmd = parser.read(line)
                cmd.execute(env)
            env.update_addr_to_next_command()
            value = env.returned_value
            env.returned_value = None
            return value

        env.context.define(self.name, _func)


class Return(Command):
    def __init__(self, expr: typing.Any):
        self.expr = expr

    def execute(self, env: Environment) -> None:
        env.returned_value = env.evaluate(self.expr)
        while True:
            block = env.blocks[-1]
            if block.kind == BlockKind.CALL:
                block.did_exit(env)
                return
            if block.kind in (
                BlockKind.IFS,
                BlockKind.IF_ELIF_ELSE,
                BlockKind.FOR,
                BlockKind.WHILE,
            ):
                env.blocks.pop()
                continue
            raise InvalidReturnError()



for i, w in enumerate(["test", "test" * 2]):
    for c in w * (i + 1):
        print(c, end=" ")
    print()

a, b = 3, 7
d = {"test": b, "test2": a}
for k in d.keys():
    print("{}: {}".format(k, d[k]), end=" // ")
print()



import typing
from .command import Command
from ..environment import Environment


class ExprStmt(Command):
    def __init__(self, expr: typing.Any):
        self.expr = expr

    def execute(self, env: Environment) -> None:
        env.evaluate(self.expr)



# Cannot define as enum because of the duplicate values
class Index:
    # command
    INDENT = 0
    OPTION = 1
    KEYWORD = 2

    # statement
    ASSIGN_LEFT = 3
    ASSIGN_RIGHT = 4

    CONDITION = 3

    FOR_VARIABLES = 3
    FOR_ITERABLE = 4

    DEF_NAME = 3
    DEF_PARAMETERS = 4

    RETURN_VALUE = 3

    CLASS_NAME = 3
    CLASS_SUPERCLASS = 4

    IMPORT_PATH = 3

    EXPR_STMT = 3

    # expression
    EXPRESSION_KEYWORD = 0

    VAR_NAME = 1

    ATTR_OBJECT = 1
    ATTR_NAME = 2

    SUBSCRIPT_OBJECT = 1
    SUBSCRIPT_INDEX = 2
    SUBSCRIPT_SLICE_START = 2
    SUBSCRIPT_SLICE_STOP = 3

    CALL_CALLEE = 1
    CALL_ARGS = 2

    KWARG_NAME = 1
    KWARG_VALUE = 2

    LEFT_OPERAND = 1
    RIGHT_OPERAND = 2

    UNARY_OPERAND = 1



import enum
import typing
import json

from .address import Address
from .command.command import Command
from .command.pass_stmt import End
from .element import Element
from .environment import Environment
from .index import Index
from .keyword import Keyword
from .label import FunctionCalled, InputCalled
from .parser import Parser


class RuntimeResult(enum.Enum):
    TERMINATED = 0
    EXECUTED = 1
    BREAKPOINT = 2
    EXCEPTION = 3
    PAUSED = 4


class Runtime:
    def __init__(self, code: typing.Union[str, list]):
        commands: list
        if isinstance(code, str):
            commands = json.loads(code)
        else:
            commands = code
        self.env = Environment(commands)
        self.breakpoints = set()
        self.parser = Parser()
        self._calls = []
        self._inputcmd: typing.Optional[Command] = None

    def resume(self, inputstr: str) -> RuntimeResult:
        """Resumes the runtime after an input() call."""
        self.env.returned_value = inputstr
        cmd: Command = self._inputcmd  # type: ignore
        self._inputcmd = None
        try:
            cmd.execute(self.env)
            self.env.update_addr_to_next_command()
        except FunctionCalled:
            caller_addr = self.env.addr.clone()
            self._calls.append(CallingCommand(caller_addr, cmd))
        except InputCalled:
            self._inputcmd = cmd
            return RuntimeResult.PAUSED
        except:
            raise
        return RuntimeResult.EXECUTED

    def run(self) -> RuntimeResult:
        while True:
            result = self.step()
            if result == RuntimeResult.EXECUTED:
                continue
            else:
                return result

    def step(self) -> RuntimeResult:
        last_index = len(self.env.code) - 1
        if self.env.addr.indent == 0:
            return RuntimeResult.TERMINATED
        if self.env.addr.line >= last_index:
            return RuntimeResult.TERMINATED

        line = self.env.code[self.env.addr.line]
        cmd = self.parser.read(line)

        caller_addr = self.env.addr.clone()
        if len(self._calls) > 0 and caller_addr.is_at(self._calls[-1].addr):
            last_called_cmd = self._calls.pop()
            cmd = last_called_cmd.cmd

        try:
            cmd.execute(self.env)
        except FunctionCalled:
            self._calls.append(CallingCommand(caller_addr, cmd))
        except InputCalled:
            self._inputcmd = cmd
            return RuntimeResult.PAUSED
        except:
            raise

        if isinstance(cmd, End):
            return RuntimeResult.TERMINATED

        self.env.update_addr_to_next_command()
        next_line = self.env.code[self.env.addr.line]
        kwd = Keyword(next_line[Index.KEYWORD])
        while kwd in (Keyword.COMMENT, Keyword.IFS):
            cmd = self.parser.read(next_line)
            cmd.execute(self.env)
            self.env.update_addr_to_next_command()
            next_line = self.env.code[self.env.addr.line]
            kwd = Keyword(next_line[Index.KEYWORD])

        if self.env.addr.line in self.breakpoints:
            return RuntimeResult.BREAKPOINT

        return RuntimeResult.EXECUTED


class CallingCommand:
    def __init__(self, addr: Address, cmd: Command):
        self.addr = addr
        self.cmd = cmd



class Address:
    def __init__(self, indent: int, line: int, calls=0):
        self.indent = indent
        self.line = line
        self.calls = calls

    def clone(self):
        return Address(self.indent, self.line, self.calls)

    def is_at(self, addr: "Address"):
        return (
            self.indent == addr.indent
            and self.line == addr.line
            and self.calls == addr.calls
        )

    def jump(self, addr: "Address"):
        self.indent = addr.indent
        self.line = addr.line
        # calls is not changed

    def shift(self, delta_indent: int, delta_line=0):
        self.indent += delta_indent
        self.line += delta_line
        # calls is not changed



import typing
from ..environment import Environment
import abc
from ..error import (
    OutOfRangeError,
    KeyNotContainedError,
    AssignmentNotSupportedError,
)


class Assignable(abc.ABC):
    @abc.abstractmethod
    def assign(self, value: typing.Any, env: Environment) -> None:
        pass

    @abc.abstractmethod
    def evaluate(self, env: Environment) -> typing.Any:
        pass


class Variable(Assignable):
    def __init__(self, name: str) -> None:
        self.name = name

    def assign(self, value: typing.Any, env: Environment) -> None:
        env.context.define(self.name, value)

    def evaluate(self, env: Environment) -> typing.Any:
        return env.context.lookup(self.name)

    def __str__(self) -> str:
        return self.name


class Attribute(Assignable):
    def __init__(
        self, obj: typing.Union[str, Assignable], properties: list[str]
    ) -> None:
        self.obj = obj
        self.properties = properties

    def assign(self, value: typing.Any, env: Environment) -> None:
        target = self._lookup(env)
        for prop in self.properties[:-1]:
            target = getattr(target, prop)
        setattr(target, self.properties[-1], value)

    def evaluate(self, env: Environment) -> typing.Any:
        target = self._lookup(env)
        for prop in self.properties:
            target = getattr(target, prop)
        return target

    def _lookup(self, env: Environment) -> typing.Any:
        if isinstance(self.obj, Assignable):
            return self.obj.evaluate(env)
        return self.obj

    def __str__(self) -> str:
        return f'{self.obj}.{".".join(self.properties)}'


KeyType = typing.Union[int, str, Assignable]
KeyIndex = typing.Union[int, Variable]


class Subscript(Assignable):
    def __init__(
        self,
        ref: Assignable,
        key: typing.Optional[KeyType],
        start: typing.Optional[KeyIndex] = None,
        stop: typing.Optional[KeyIndex] = None,
    ) -> None:
        self.ref = ref
        self.key = key
        self.start = start
        self.stop = stop

    def assign(self, value: typing.Any, env: Environment) -> None:
        obj: typing.Any = self.ref.evaluate(env)
        key: typing.Any = env.evaluate(self.key)
        start: typing.Optional[int] = env.evaluate(self.start)
        stop: typing.Optional[int] = env.evaluate(self.stop)

        if key is not None:
            try:
                obj[key] = value
                return
            except IndexError:
                raise OutOfRangeError(str(self.ref), str(key))
            except TypeError:
                raise AssignmentNotSupportedError(str(self.ref))
        if start is None:
            if stop is None:
                obj[:] = value
                return
            else:
                obj[:stop] = value
                return
        else:
            if stop is None:
                obj[start:] = value
                return
            else:
                obj[start:stop] = value
                return

    def evaluate(self, env: Environment) -> typing.Any:
        obj: typing.Any = self.ref.evaluate(env)
        key: typing.Any = env.evaluate(self.key)
        start: typing.Optional[int] = env.evaluate(self.start)
        stop: typing.Optional[int] = env.evaluate(self.stop)

        if key is not None:
            try:
                return obj[key]
            except IndexError:
                raise OutOfRangeError(str(self.ref), str(key))
            except KeyError:
                raise KeyNotContainedError(str(self.ref), str(key))
        if start is None:
            if stop is None:
                return obj[:]
            else:
                return obj[:stop]
        else:
            if stop is None:
                return obj[start:]
            else:
                return obj[start:stop]

    def __str__(self) -> str:
        return f"{self.ref}[{self.key}]"



import unittest
from contextlib import redirect_stdout
import io
import os

import sys
import json
import sys

sys.path.append("../src")

from calciumlang.runtime import Runtime

from calciumlang.tools.converter import convert

dir_name = None
file_names = None
if len(sys.argv) > 1:  # sys.argv[0] is the name of this file
    file_names = sys.argv[1:]
else:
    dir_name = "test_src"


def run_calcium(filepath):
    with open(filepath) as fin:
        text = fin.read()
        json_text = convert(text)
        code = json.loads(json_text)
        runtime = Runtime(code)
        runtime.run()


if __name__ == "__main__":
    suite = unittest.TestSuite()
    files = []
    if dir_name is not None:
        files = os.listdir(dir_name)
    if file_names is not None:
        files = file_names
    for filename in files:
        if not filename.endswith(".py"):
            continue

        def make_test(filepath):
            def test_file(self):
                with io.StringIO() as raw_out, io.StringIO() as calcium_out:
                    with redirect_stdout(raw_out):
                        with open(filepath) as fin:
                            exec(fin.read(), {})
                    with redirect_stdout(calcium_out):
                        run_calcium(filepath)
                    self.assertEqual(raw_out.getvalue(), calcium_out.getvalue())

            return test_file

        testname = filename[:-3]
        if dir_name is not None:
            filepath = os.path.join(os.getcwd(), dir_name, filename)
        else:
            filepath = os.path.join(os.getcwd(), filename)
        methodname = "test_{}".format(filename)
        testcase = type(
            testname, (unittest.TestCase,), {methodname: make_test(filepath)}
        )
        suite.addTest(testcase(methodName=methodname))
    unittest.TextTestRunner().run(suite)



import typing
from .address import Address
from .block_result import BlockResult
from .element import Element
from .index import Index
from .namespace import GlobalScope, Namespace


class NextLineCalculation:
    def __init__(self, block_result: BlockResult, processing_line: int):
        self.block_result = block_result
        self.processing_line = processing_line


class Environment:
    def __init__(self, code: list):
        from .block import Block

        self.code: list[list[Element]] = code
        self.addr = Address(1, 0)
        self.blocks: list[Block] = []
        self.callstack: list[Namespace] = []

        self.global_context = GlobalScope(None, {})
        self.context: Namespace = self.global_context

        self.prompt = ""
        self.returned_value: typing.Any = None

    def evaluate(self, obj: typing.Any) -> typing.Any:
        # to avoid circular imports, these must write here
        from .expression.assignable import Assignable
        from .expression.call import Call
        from .expression.operator import UnaryOperator, BinaryOperator

        if isinstance(obj, Assignable):
            return obj.evaluate(self)
        if isinstance(obj, BinaryOperator):
            return obj.evaluate(self)
        if isinstance(obj, Call):
            return obj.evaluate(self)
        if isinstance(obj, UnaryOperator):
            return obj.evaluate(self)
        if isinstance(obj, list):
            return [self.evaluate(elem) for elem in obj]
        if isinstance(obj, dict):
            return {key: self.evaluate(value) for key, value in obj.items()}
        if isinstance(obj, tuple):
            return tuple(self.evaluate(elem) for elem in obj)
        if isinstance(obj, set):
            return {self.evaluate(elem) for elem in obj}
        return obj

    def update_addr_to_next_command(self) -> None:
        next_line_index = 0
        while True:
            next_line_index = self.addr.line + 1
            calculating_next_line = self._pop_blocks(next_line_index)
            next_line_index = calculating_next_line.processing_line
            if calculating_next_line.block_result == BlockResult.SHIFT:
                break
        self.addr.line = next_line_index

    def _pop_blocks(self, line_index: int) -> NextLineCalculation:
        working_line = line_index
        while True:
            next_line: list[Element] = self.code[working_line]
            next_indent: int = next_line[Index.INDENT]  # type: ignore
            delta_indent = self.addr.indent - next_indent
            if delta_indent < 0:
                working_line += 1
                continue
            for _ in range(delta_indent):
                block = self.blocks[-1]
                block_result = block.did_exit(self)
                if block_result == BlockResult.JUMP:
                    return NextLineCalculation(BlockResult.JUMP, working_line)
            return NextLineCalculation(BlockResult.SHIFT, working_line)



import enum


class Keyword(enum.Enum):
    # assignment
    ASSIGN = "="
    COMPOUND_ADD = "+="
    COMPOUND_SUBTRACT = "-="
    COMPOUND_MULTIPLY = "*="

    # operators
    ADD = "+"
    SUBTRACT = "-"
    MULTIPLY = "*"
    DIVIDE = "/"
    MODULO = "%"
    FLOOR_DIVIDE = "//"
    POWER = "**"

    EQUAL = "=="
    NOT_EQUAL = "!="
    LESS_THAN = "<"
    LESS_THAN_EQUAL = "<="
    GREATER_THAN = ">"
    GREATER_THAN_EQUAL = ">="

    AND = "and"
    OR = "or"
    NOT = "not"

    IN = "in"

    NEGATIVE = "-_"

    # statements
    IFS = "ifs"
    IF = "if"
    ELIF = "elif"
    ELSE = "else"

    FOR = "for"
    WHILE = "while"
    BREAK = "break"
    CONTINUE = "continue"

    DEF = "def"
    RETURN = "return"

    CLASS = "class"

    IMPORT = "import"

    EXPR_STMT = "expr"

    PASS = "pass"

    # expressions
    VARIABLE = "var"
    ATTRIBUTE = "attr"
    SUBSCRIPT = "sub"
    CALL = "call"

    # syntax
    COMMA = ","
    KWARG = "kwarg"
    TUPLE = "tuple"

    COMMENT = "#"
    END = "end"



import typing
from .command import Command
from ..block import Block, BlockKind
from ..block_result import BlockResult
from ..environment import Environment
from ..expression.assignable import Variable, Attribute
from ..namespace import ClassScope


class Class(Command):
    def __init__(self, name: str, superclass: typing.Union[Variable, Attribute, None]):
        self.name = name
        self.superclass = superclass

    def execute(self, env: Environment) -> None:
        if self.superclass is None:
            superclass = object
        else:
            superclass = self.superclass.evaluate(env)

        def enter(env: Environment) -> bool:
            parent_scope = env.context.find_nesting_scope()
            env.context = ClassScope(parent_scope, {})
            return True

        previous_context = env.context

        def exit(env: Environment) -> BlockResult:
            if not isinstance(env.context, ClassScope):
                raise RuntimeError("context is not ClassScope")
            attributes = env.context.create_attributes()
            env.context = previous_context
            classtype = type(self.name, (superclass,), attributes)
            env.context.define(self.name, classtype)
            env.addr.shift(-1)
            return BlockResult.SHIFT

        block = Block(BlockKind.CLASS, env.addr, enter, exit)
        block.will_enter(env)
